@@ -3,13 +3,15 @@ use warnings;
 
 package JMAP::Tester;
 # ABSTRACT: a JMAP client made for testing JMAP servers
-$JMAP::Tester::VERSION = '0.005';
+$JMAP::Tester::VERSION = '0.006';
 use Moo;
 
 use Encode qw(encode_utf8);
 use HTTP::Request;
 use JMAP::Tester::Response;
+use JMAP::Tester::Result::Auth;
 use JMAP::Tester::Result::Failure;
+use JMAP::Tester::Result::Upload;
 use Safe::Isa;
 use URI;
 
@@ -29,7 +31,7 @@ use URI;
 #pod You use the test client like this:
 #pod
 #pod   my $jtest = JMAP::Tester->new({
-#pod     jmap_uri => 'https://jmap.local/account/123',
+#pod     api_uri => 'https://jmap.local/account/123',
 #pod   });
 #pod
 #pod   my $response = $jtest->request([
@@ -88,18 +90,21 @@ has _json_typist => (
   },
 );
 
-has jmap_uri => (
-  is => 'ro',
-  isa => sub {
-    die "provided jmap_uri is not a URI object" unless $_[0]->$_isa('URI');
-  },
-  coerce => sub {
-    return $_[0] if $_[0]->$_isa('URI');
-    die "can't use reference as a URI" if ref $_[0];
-    URI->new($_[0]);
-  },
-  required => 1,
-);
+for my $type (qw(api authentication download upload)) {
+  has "$type\_uri" => (
+    is => 'rw',
+    isa => sub {
+      die "provided $type\_uri is not a URI object" unless $_[0]->$_isa('URI');
+    },
+    coerce => sub {
+      return $_[0] if $_[0]->$_isa('URI');
+      die "can't use reference as a URI" if ref $_[0];
+      URI->new($_[0]);
+    },
+    predicate => "has_$type\_uri",
+    clearer   => "clear_$type\_uri",
+  );
+}
 
 has ua => (
   is   => 'ro',
@@ -124,8 +129,8 @@ sub _set_cookie {
     $name,
     $value,
     '/',
-    $self->jmap_uri->host,
-    $self->jmap_uri->port,
+    $self->api_uri->host,
+    $self->api_uri->port,
     0,
     0,
     86400,
@@ -190,6 +195,9 @@ has default_arguments => (
 sub request {
   my ($self, $calls) = @_;
 
+  Carp::confess("can't perform request: no api_uri configured")
+    unless $self->has_api_uri;
+
   state $ident = 'a';
   my %seen;
   my @suffixed;
@@ -228,7 +236,7 @@ sub request {
   my $json = $self->json_encode(\@suffixed);
 
   my $post = HTTP::Request->new(
-    POST => $self->jmap_uri->as_string,
+    POST => $self->api_uri->as_string,
     [
       'Content-Type' => 'application/json',
     ],
@@ -256,6 +264,197 @@ sub request {
   });
 }
 
+#pod =method upload
+#pod
+#pod   my $result = $tester->upload($mime_type, $blob_ref, \%arg);
+#pod
+#pod This uploads the given blob, which should be given as a reference to a string.
+#pod
+#pod The return value will either be a L<failure
+#pod object|JMAP::Tester::Result::Failure> or an L<upload
+#pod result|JMAP::Tester::Result::Upload>.
+#pod
+#pod =cut
+
+sub upload {
+  my ($self, $mime_type, $blob_ref) = @_;
+  # TODO: support blob as handle or sub -- rjbs, 2016-11-17
+
+  Carp::confess("can't upload without upload_uri")
+    unless $self->upload_uri;
+
+  my $res = $self->ua->post(
+    $self->upload_uri,
+    Content => $$blob_ref,
+    'Content-Type' => $mime_type,
+  );
+
+  unless ($res->is_success) {
+    return JMAP::Tester::Result::Failure->new({
+      http_response => $res,
+    });
+  }
+
+  return JMAP::Tester::Result::Upload->new({
+    http_response => $res,
+    payload       => $self->apply_json_types(
+      $self->json_decode( $res->decoded_content )
+    ),
+  });
+}
+
+#pod =method download
+#pod
+#pod   my $result = $tester->download(\%arg);
+#pod
+#pod Valid arguments are:
+#pod
+#pod   blobId    - the blob to download (required)
+#pod   accountId - the account for which we're downloading (required)
+#pod   name      - the name we want the server to provide back (default: "download")
+#pod
+#pod
+#pod The return value will either be a L<failure
+#pod object|JMAP::Tester::Result::Failure> or an L<upload
+#pod result|JMAP::Tester::Result::Download>.
+#pod
+#pod =cut
+
+sub download {
+  my ($self, $arg) = @_;
+
+  Carp::confess("can't download without download") unless $self->download_uri;
+
+  for my $name (qw(blobId accountId)) {
+    Carp::confess("missing required parameter $name")
+      unless defined $arg->{$name};
+  }
+
+  my $uri = $self->download_uri;
+  for my $var (qw(blobId accountId name)) {
+    my $value = $arg->{$var} // $var; # only "name" could be undef
+    $uri =~ s/\{$var\}/$arg->{$var}/g;
+  }
+
+  my $res = $self->ua->get($uri);
+
+  unless ($res->is_success) {
+    return JMAP::Tester::Result::Failure->new({
+      http_response => $res,
+    });
+  }
+
+  return JMAP::Tester::Result::Download->new({
+    http_response => $res,
+  });
+}
+
+#pod =method simple_auth
+#pod
+#pod   my $auth_struct = $tester->simple_auth($username, $password);
+#pod
+#pod =cut
+
+sub simple_auth {
+  my ($self, $username, $password) = @_;
+
+  # This is fatal, not a failure return, because it reflects the user screwing
+  # up, not a possible JMAP-related condition. -- rjbs, 2016-11-17
+  Carp::confess("can't simple_auth: no authentication_uri configured")
+    unless $self->has_authentication_uri;
+
+  my $start_json = $self->json_encode({
+    username      => $username,
+    clientName    => (ref $self),
+    clientVersion => $self->VERSION // '0',
+    deviceName    => 'JMAP Testing Client',
+  });
+
+  my $start_res = HTTP::Request->new(
+    POST => $self->authentication_uri->as_string,
+    [
+      'Content-Type' => 'application/json',
+    ],
+    $start_json,
+  );
+
+  unless ($start_res->code == 200) {
+    return JMAP::Tester::Result::Failure->new({
+      ident         => 'failure in auth phase 1',
+      http_response => $start_res,
+    });
+  }
+
+  my $start_reply = $self->json_decode( $start_res->decoded_content );
+
+  unless (grep {; $_->{type} eq 'password' } @{ $start_reply->{methods} }) {
+    return JMAP::Tester::Result::Failure->new({
+      ident         => "password is not an authentication method",
+      http_response => $start_res,
+    });
+  }
+
+  my $next_json = $self->json_encode({
+    loginId => $start_reply->{loginId},
+    type    => 'password',
+    value   => $password,
+  });
+
+  my $next_res = HTTP::Request->new(
+    POST => $self->authentication_uri->as_string,
+    [
+      'Content-Type' => 'application/json',
+    ],
+    $next_json,
+  );
+
+  unless ($next_res->code == 201) {
+    return JMAP::Tester::Result::Failure->new({
+      ident         => 'failure in auth phase 2',
+      http_response => $next_res,
+    });
+  }
+
+  my $auth_struct = $self->json_decode( $next_res->decoded_content );
+
+  $self->{access_token} = $auth_struct->{accessToken};
+
+  $self->ua->default_header(
+    'Authorization' => "Bearer $auth_struct->{accessToken}"
+  );
+
+  my $auth = JMAP::Tester::Result::Auth->new({
+    http_response => $next_res,
+    auth_struct   => $auth_struct,
+  });
+
+  $self->_update_uris_from_auth($auth);
+
+  return $auth;
+}
+
+sub _update_uris_from_auth {
+  my ($self, $auth) = @_;
+
+  # It's not crazy to think that we'd also try to pull the primary accountId
+  # out of the accounts in the auth struct, but I don't think there's a lot to
+  # gain by doing that yet.  Maybe later we'd use it to set the default
+  # X-JMAP-AccountId or other things, but I think there are too many open
+  # questions.  I'm leaving it out on purpose for now. -- rjbs, 2016-11-18
+
+  for my $type (qw(api download upload)) {
+    if (defined (my $uri = $auth->auth_struct->{"${type}Url"})) {
+      my $setter = "$type\_uri";
+      $self->$setter($uri);
+    } else {
+      my $clearer = "clear_$type\_uri";
+      $self->$clearer;
+    }
+  }
+
+  return;
+}
+
 1;
 
 __END__
@@ -270,7 +469,7 @@ JMAP::Tester - a JMAP client made for testing JMAP servers
 
 =head1 VERSION
 
-version 0.005
+version 0.006
 
 =head1 OVERVIEW
 
@@ -288,7 +487,7 @@ L<Paragraphs|JMAP::Tester::Response::Paragraph>.
 You use the test client like this:
 
   my $jtest = JMAP::Tester->new({
-    jmap_uri => 'https://jmap.local/account/123',
+    api_uri => 'https://jmap.local/account/123',
   });
 
   my $response = $jtest->request([
@@ -370,6 +569,34 @@ at least, failures are L<JMAP::Tester::Result::Failure> objects.  More refined
 failure objects may exist in the future.  Successful requests return
 L<JMAP::Tester::Response> objects.
 
+=head2 upload
+
+  my $result = $tester->upload($mime_type, $blob_ref, \%arg);
+
+This uploads the given blob, which should be given as a reference to a string.
+
+The return value will either be a L<failure
+object|JMAP::Tester::Result::Failure> or an L<upload
+result|JMAP::Tester::Result::Upload>.
+
+=head2 download
+
+  my $result = $tester->download(\%arg);
+
+Valid arguments are:
+
+  blobId    - the blob to download (required)
+  accountId - the account for which we're downloading (required)
+  name      - the name we want the server to provide back (default: "download")
+
+The return value will either be a L<failure
+object|JMAP::Tester::Result::Failure> or an L<upload
+result|JMAP::Tester::Result::Download>.
+
+=head2 simple_auth
+
+  my $auth_struct = $tester->simple_auth($username, $password);
+
 =head1 AUTHOR
 
 Ricardo SIGNES <rjbs@cpan.org>
